/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
hendrik.lipka@gmx.de
*/

package de.hendriklipka.ldse.common;

import java.util.Vector;

import de.hendriklipka.ldse.common.Log;
/**
  Stellt ein Lock-Objekt fuer einen Reader-Writer-Lock zur Verfgung.
  Auf eine solche Ressource knnen beliebig viele lesende Zugriffe parallel
  erfolgen, aber schreibende Zugriffe mssen alleine stattfinden.
  Dabei kann ein Thread, der einen schreibenden Zugriff ausfhrt, auch lesende
  Zugriffe und andere schreibende Zugriffe ausfhren.
  Aus einem Block mit ReaderLock darf kein WriterLock angefordert werden
  Benutzung:
  <CODE>
  ReaderWriterLock lock=new ReaderWriterLock();
  // lesender Zugriff
  lock.requestRead();
  try
  {
  }
  finally
  {
    lock.finishedRead();
  }
  // schreibender Zugriff erfolgt quivalent
*/
public class ReaderWriterLock
{
  protected int _writeCount=0; // schreibt gerade jemand?
  protected int _readCount=0; // wieviele Threads lesen gerade?
  protected int _readWaiting=0; // wieviele warten aufs Lesen?
  protected Thread _writeOwner=null; // wer schreibt gerade?
  protected Vector _writeWaiting=new Vector(5,5); // wer warten aus Schreiben?
  protected Object _writeLock=new Object(); // Lock-Objekt fuer _isWriting & _writeWaiting & _writeOwner
  protected Object _readLock=new Object(); // Lock-Objekt fuer _readCount
  protected Object _waitForWrite=new Object(); // 
  protected Object _waitForRead=new Object(); // Lock fuer _readWaiting
  
  /**
    Lesezugriff anfordern
    wird gewaehrt, wenn keine Schreibzugriff stattfindet oder dieser
    vom aktuellen Thread durchgefuehrt wird und auch niemand auf das Schreiben wartet
    (damit Writer nicht auf ewig gesperrt werden koennen)
  */
  public void requestRead()
  {
    Log.log("lock","trace",1,"requestRead: "+Thread.currentThread().getName());
    // Lock holen
    synchronized(_writeLock)
    {
      // Schreibt jemand fremdes?
      if (((0==_writeWaiting.size())&&(0==_writeCount))||(Thread.currentThread()==_writeOwner))
      {
        synchronized(_readLock)
        {
          // Zugriff holen
          _readCount++;
          Log.log("lock","trace",1,"requestRead gained: "+Thread.currentThread().getName()+"\n readCount="+_readCount);
          return;
        }
      }
    }
    // Warten...
    synchronized(_waitForRead)
    {
      // vorher Doppelpruefung, falls der Schreibzugriff gerade eben beendet wurde...
      if (((0==_writeWaiting.size())&&(0==_writeCount))||(Thread.currentThread()==_writeOwner))
      {
        synchronized(_readLock)
        {
          _readCount++;
          Log.log("lock","trace",1,"requestRead gained: "+Thread.currentThread().getName()+"\n readCount="+_readCount);
          return;
        }
      }
      _readWaiting++;
      try
      {
        _waitForRead.wait();
      }
      catch (InterruptedException ex) {}
      // hier wurden wir aufgeweckt, also bekommen wir den Lock...
      synchronized(_readLock)
      {
        _readCount++;
        Log.log("lock","trace",1,"requestRead gained: "+Thread.currentThread().getName()+"\n readCount="+_readCount);
      }
      _readWaiting--;
    }
  }
  
  /**
    Lesezugriff auf die Ressource beenden
  */
  public void finishedRead()
  {
    Log.log("lock","trace",1,"finishedRead: "+Thread.currentThread().getName());
    // Lock holen
    synchronized(_readLock)
    {
      _readCount--;
      // War das der letzte lesende Zugriff?
      if (0==_readCount)
      {
        synchronized(_writeLock)
        {
          // wenn jemand auf das Schreiben wartet, wird er benachrichtigt
          // es muss kontrolliert werden, ob nicht auch ein Schreibzugriff besteht...
          // wenn der Aufruf aus einem Writer heraus erfolgt, ist _writeCount=1
          if ((0==_writeCount)&&(0!=_writeWaiting.size()))
          {
            Object o=_writeWaiting.firstElement();
            synchronized(o)
            {
                // Das wird vom geradde aufgeweckten Thread selbst vorgenommen!!
//              _writeWaiting.removeElementAt(0);
              o.notify();
              Log.log("lock","trace",1,"finishedRead does writeWaiting: "+Thread.currentThread().getName());
            }
          }
        }
      }
    }
  }
  
  /**
    Fordert exklusiven Schreib- und Lesezugriff auf die Ressource an.
    Kehrt erst zurck, wenn dieser Zugriff gewhrleistet wird.
    requestRead() und requestWrite() knnen rekursiv aufgerufen werden
  */
  public void requestWrite()
  {
    Log.log("lock","trace",1,"requestWrite: "+Thread.currentThread().getName());
    // Lock erzeugen
    synchronized(_writeLock)
    {
      // Schreibt sonst noch jemand?
      if ((0==_writeCount)||(Thread.currentThread()==_writeOwner))
      {
        // Liest noch jemand?
        synchronized(_readLock)
        {
          if (0==_readCount)
          { // Nein -> Schreiben also wir
            _writeOwner=Thread.currentThread();
            _writeCount++;
            Log.log("lock","trace",1,"requestWrite gained: "+Thread.currentThread().getName());
            return;
          }
        }
      }
    }
    // Lock wurde nicht erlaubt
    // Eigenes Lock-Objekt erzeugen
    Object o=new Object();
    synchronized(o)
    {
      // Lock-Objekt in die Warteliste eintragen
      _writeWaiting.addElement(o);
      // Doppel-Check: nochmal nachschauen, ob ich vielleicht jetzt schreiben kann
      // andere Threads koennen jetzt kein Notify auf das Lock-Objekt machen,
      // da wir hier den Lock darauf haben
      // wenn wir hier den Schreibzugriff bekommen, macht der andere Thread hinter ein
      // Notify und ist dann fertig
      synchronized(_writeLock)
      {
        if ((0==_writeCount)||(Thread.currentThread()==_writeOwner))
        {
          synchronized(_readLock)
          {
            if (0==_readCount)
            {
              _writeOwner=Thread.currentThread();
              _writeCount++;
              _writeWaiting.removeElement(o);
              Log.log("lock","trace",1,"requestWrite gained: "+Thread.currentThread().getName());
              return;
            }
          }
        }
      }
      try
      {
        o.wait();
      }
      catch (InterruptedException ex) {}
      // hier wurden wir aufgeweckt, also bekommen wir den Lock...
      _writeOwner=Thread.currentThread();
      _writeCount++;
      _writeWaiting.removeElement(o); // erst jetzt koennte jemand anderes Lock bekommen
                                      // aber da steht _writeCount schon auf 1...
      Log.log("lock","trace",1,"requestWrite gained: "+Thread.currentThread().getName());
    }
  }
  
  /**
    exklusiven Schreibzugriff auf die Ressource freigeben
  */
  public void finishedWrite()
  {
    Log.log("lock","trace",1,"finishedWrite: "+Thread.currentThread().getName());
    // Lock holen
    synchronized(_writeLock)
    {
      // Zugriff aufgeben
      _writeCount--;
      // wenn wir noch Zugriff haben, war es das
      if (0!=_writeCount)
        return;
      _writeOwner=null;
      // nachsehen, ob es Reader zum Starten gibt
      synchronized(_waitForRead)
      {
        if(0!=_readWaiting)
        { // ja -> benachrichtigen und fertig
          _waitForRead.notifyAll();
          Log.log("lock","trace",1,"finishedWrite does readWaiting: "+Thread.currentThread().getName());
          return;
        }
      }
      // keine Reader, also sehen wir nach Writers
      if (0!=_writeWaiting.size())
      {
        Object o=_writeWaiting.firstElement();
        // in diesen Block kommen wir erst, wenn niemand anderes gerade
        // sich eintragen will
        synchronized(o)
        {
          // wartenden Thread benachrichtigen
          // macht der geweckte Thread selber...
//          _writeWaiting.removeElementAt(0);
          Log.log("lock","trace",1,"finishedWrite does writeWaiting: "+Thread.currentThread().getName());
          o.notify();
        }
      }
    }
  }
}